#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <vector>
#include <iostream>
#include <fstream>

#define	COUT	std::cout
#define ENDL	std::endl
#define	MVECT	std::vector<Message>
#define	IVECT	std::vector<int>

#include "Pipe.h"
#include "Message.h"


//#define DIMENSION	5

IVECT drops;
IVECT delays;

pthread_mutex_t netforwardmutex = PTHREAD_MUTEX_INITIALIZER;
sem_t netforwardsem;
pthread_mutex_t netbackwardmutex = PTHREAD_MUTEX_INITIALIZER;
sem_t netbackwardsem;

sem_t timesem;

MVECT forwardMessages;
MVECT backwardMessages;

long wallclock = 0;
int senderArrived = 0;
int receiverArrived = 0;

ReceivePipe receiverFromSource("NETRECFMSRC");
SendPipe senderToDetination("NETSNDTODEST");

ReceivePipe receiverFromDestination("NETRECFMDEST");
SendPipe senderToSource("NETSNDTOSRC");

void timer(int signum)
{
	wallclock++;
	
	sem_post (&timesem);		
}

void * time_cleaner(void * parameter)
{
	while(true)
	{
		sem_wait (&timesem);
		
		pthread_mutex_lock(&netforwardmutex);//{
		senderArrived = 0;
		pthread_mutex_unlock(&netforwardmutex);//}
		
		pthread_mutex_lock(&netbackwardmutex);//{
		receiverArrived = 0;
		pthread_mutex_unlock(&netbackwardmutex);//}		
	}
	
	return NULL;
}

void * sender_to_destination_thread(void * parameter)
{
	while(true)
	{
		sem_wait (&netforwardsem);
	
		Message m;
		bool found = false;
	
		pthread_mutex_lock(&netforwardmutex);//{ 1

		for(int i = 0; i < forwardMessages.size(); i++)
		{
			found = false;
			
			if(forwardMessages[i].getTimeStamp() < wallclock)
			{
				found = true;
				m = forwardMessages[i];
				
				forwardMessages.erase(forwardMessages.begin() + i);
				i--;
				
				pthread_mutex_unlock(&netforwardmutex);//} 1, 2

				COUT << "Sent to destination in time " << wallclock << ' ' << m.toString().c_str() << ENDL;
							
				senderToDetination.send(m);
				sem_wait (&netforwardsem);

				pthread_mutex_lock(&netforwardmutex);//{ 2			
			}
		}
	
		pthread_mutex_unlock(&netforwardmutex);//}
		
		//An extra wait has happend and nothing has been sent
		if(!found)
			sem_post (&netforwardsem);		
	}
	
	return NULL;
}

void * sender_to_source_thread(void * parameter)
{
	while(true)
	{
		sem_wait (&netbackwardsem);
	
		Message m;
		bool found = false;
	
		pthread_mutex_lock(&netbackwardmutex);//{ 1
		
		for(int i = 0; i < backwardMessages.size(); i++)
		{
			found = false;
			
			if(backwardMessages[i].getTimeStamp() < wallclock)
			{
				found = true;
				m = backwardMessages[i];
				
				backwardMessages.erase(backwardMessages.begin() + i);
				i--;
				
				pthread_mutex_unlock(&netbackwardmutex);//} 1, 2
			
				COUT << "Sent to source in time " << wallclock << ' ' << m.toString().c_str() << ENDL;
							
				senderToSource.send(m);
				sem_wait (&netbackwardsem);
				
				pthread_mutex_lock(&netbackwardmutex);//{ 2			
			}
		}
	
		pthread_mutex_unlock(&netbackwardmutex);//}
		
		//An extra wait has happend and nothing has been sent
		if(!found)
			sem_post (&netbackwardsem);		
	}
	
	return NULL;
}

void * receiver_from_source_thread(void * parameter)
{
	while(true)
	{
		Message m = receiverFromSource.receive();	
	
		pthread_mutex_lock(&netforwardmutex);//{
			
		senderArrived++;
		
		COUT << "Received from source in time " << wallclock << ' ' << m.toString().c_str() << ENDL;

		m.incTimeStamp( delays[m.getTimeStamp() % delays.size()]);
		
		if(senderArrived >= drops[m.getTimeStamp() % drops.size()])
		{
			COUT << "!! Dropped src " << m.toString().c_str() << ENDL;
			senderArrived = 0;
		}
		else
		{
			forwardMessages.push_back(m);
			sem_post (&netforwardsem);
		}
		
		pthread_mutex_unlock(&netforwardmutex);//}
	}

	return NULL;
}

void * receiver_from_destination_thread(void * parameter)
{
	while(true)
	{
		Message m = receiverFromDestination.receive();	
	
		pthread_mutex_lock(&netbackwardmutex);//{
			
		receiverArrived++;

		COUT << "Received from destination in time " << wallclock << ' ' << m.toString().c_str() << ENDL;

		m.incTimeStamp( delays[m.getTimeStamp() % delays.size()]);
		
		if(receiverArrived >= drops[m.getTimeStamp() % drops.size()])
		{
			COUT << "!! Dropped dest " << m.toString().c_str() << ENDL;
			receiverArrived = 0;
		}
		else
		{
			backwardMessages.push_back(m);
			sem_post (&netbackwardsem);
		}
		
		pthread_mutex_unlock(&netbackwardmutex);//}
	}

	return NULL;
}

void init()
{
	sem_init (&netforwardsem, 0, 0);
	sem_init (&netbackwardsem, 0, 0);
	sem_init (&timesem, 0, 0);
	
	signal(SIGUSR1, &timer);

	{
		SendPipe PIDSender("NETWORKTELLPID");
		
		Message m( 0, 0, 0, getpid());
		PIDSender.send(m);
		COUT << "PID: " << m.getTimeStamp() << ENDL;	
	}	
	
	{
		std::ifstream dropFile("drops.txt");
		
		if(!dropFile.good())
		{
			drops.push_back(10);
			drops.push_back(13);
			drops.push_back(12);
			drops.push_back(7);
			drops.push_back(8);
		}
		else
		{
			int i;
			dropFile >> i;
		
			while(dropFile.good())
			{
				COUT << i << ' ';
				dropFile >> i;
				drops.push_back(i);
			}
		
			COUT << ENDL;		
		}
	
		dropFile.close();
		
		std::ifstream delayFile("delays.txt");
		
		if(!delayFile.good())
		{	
			delays.push_back(1);
			delays.push_back(3);
			delays.push_back(2);
			delays.push_back(1);
			delays.push_back(1);
		}
		else
		{
			int i;
			delayFile >> i;
		
			while(delayFile.good())
			{
				COUT << i << ' ';
				delayFile >> i;
				delays.push_back(i);
			}
		
			COUT << ENDL;
		}
	
		delayFile.close();
	}
}

int main(int argc, char *argv[])
{
	init();
	
	pthread_t for_send_to_destination;
	pthread_t for_receive_from_source;
	pthread_t for_send_to_source;
	pthread_t for_receive_from_destination;
	pthread_create(&for_send_to_destination, NULL, &sender_to_destination_thread, NULL);
	pthread_create(&for_receive_from_source, NULL, &receiver_from_source_thread, NULL);
	pthread_create(&for_send_to_source, NULL, &sender_to_source_thread, NULL);
	pthread_create(&for_receive_from_destination, NULL, &receiver_from_destination_thread, NULL);

	pthread_join(for_send_to_destination, NULL);
	pthread_join(for_receive_from_source, NULL);
	pthread_join(for_send_to_source, NULL);
	pthread_join(for_receive_from_destination, NULL);

	return 0;
}

//	pthread_t for_source;
//	pthread_t for_destination;
//	pthread_create(&for_destination, NULL, &destination_thread, NULL);
//	pthread_create(&for_source, NULL, &source_thread, NULL);
//	pthread_join(for_destination, NULL);
//	pthread_join(for_source, NULL);

//void * destination_thread(void * parameter)
//{
//	while(true)
//	{
//		Message m = receiverFromDestination.receive();
//		pthread_mutex_lock(&netforwardmutex);//{
//		printf("Received from destination: %s", m.toString().c_str());
//		pthread_mutex_unlock(&netforwardmutex);//}
//		receiverArrived++;
//
//		if(receiverArrived == drops[m.getTimeStamp() % drops.size()])
//		{
//			printf("!! Dropped src %s", m.toString().c_str());
//			receiverArrived = 0;
//		}
//		else
//		{
//			senderToSource.send(m);
//		}
//		
//	}
//
//	return NULL;
//}
//
//void * source_thread(void * parameter)
//{
//	while(true)
//	{
//		Message m = receiverFromSource.receive();
//		pthread_mutex_lock(&netforwardmutex);//{
//		printf("Received from source: %s", m.toString().c_str());
//		pthread_mutex_unlock(&netforwardmutex);//}
//		senderArrived++;
//		
//		if(senderArrived == drops[m.getTimeStamp() % drops.size()])
//		{
//			printf("!! Dropped src %s", m.toString().c_str());
//			senderArrived = 0;
//		}
//		else
//		{
//			senderToDetination.send(m);
//		}
//	}
//
//	return NULL;
//}
